//
// Some of this code was copied from UDP-CPP: https://github.com/UnknownDetectionParty/UDP-CPP
//

#include "Mapping.h"

std::map<std::string, CM*> lookup;

CM *Mapping::getClass(const char *key)  {
    CM* cm = lookup.at(std::string(key));
    return cm;
}

const char *Mapping::getClassName(const char *key)  {
    return getClass(key)->name;
}

void Mapping::setup() {
    // How to define mappings:
    // --- Unobfuscated classes:
    //
    //   Creating mappings for a class:
    //
    //                SimpleID,    InternalName
    //       m = make("Set",       "java/util/Set");
    //
    //   Adding a member to the class:
    //              MappingInst,   MethodName,  InternalDescriptor,       IsStatic
    //       method(m,             "get",       "(I)Ljava/lang/Object;",  false);
    //
    // -------------------------
    //
    // base - normal jvm classes
    struct CM* m = make("List", "java.util.List");
    method(m, "get", "(I)Ljava/lang/Object;", false);
    method(m, "toArray", "()[Ljava/lang/Object;", false);
    method(m, "size", "()I", false);
    m = make("Set", "java.util.Set");
    method(m, "toArray", "()[Ljava/lang/Object;", false);
    method(m, "size", "()I", false);
    m = make("System", "java.lang.System");
    field(m, "out", "Ljava/io/PrintStream;", true);
    m = make("PrintStream", "java.io.PrintStream");
    method(m, "println", "(Ljava/lang/String;)V", false);
    m = make("Mouse", "org.lwjgl.input.Mouse");
    method(m, "isButtonDown", "(I)Z", true);

    // How to define mappings:
    // --- Obfuscated classes:
    //
    //   Creating mappings for a class:
    //
    //                SimpleID,    ObfuscatedName
    //       m = make("Entity",    "v");
    //
    //   Adding a member to the class:
    //              MappingInst,   SimpleID, ObfuscatedName, ObfuscatedDescriptor,  IsStatic
    //       method(m,             "getID",  "S",            "()I",                 false);
    //
    // -----------------------------------
    //
    // obfuscated minecraft classes - 1.8.9 forge
    m = make("Entity", "net.minecraft.entity.Entity");
    field(m, "x", "field_70165_t", "D", false);
    field(m, "lastTickPosX", "field_70142_S", "D", false);
    field(m, "y", "field_70163_u", "D", false);
    field(m, "z", "field_70161_v", "D", false);
    field(m, "lastTickPosZ", "field_70136_U", "D", false);
    method(m, "getEyeHeight", "func_70047_e", "()F", false);
    method(m, "getID", "func_145782_y", "()I", false);
    method(m, "getName", "func_70005_c_", "()Ljava/lang/String;", false);
    m = make("WorldClient", "net.minecraft.client.multiplayer.WorldClient");
    field(m, "entities", "field_72996_f", "Ljava/util/List;", false);
    field(m, "players", "field_73010_i", "Ljava/util/List;", false);
    method(m, "setTime", "func_72877_b", "(J)V", false);
    m = make("PlayerSP", "net.minecraft.client.entity.EntityPlayerSP");
    field(m, "x", "field_70165_t", "D", false);
    field(m, "y", "field_70163_u", "D", false);
    field(m, "z", "field_70161_v", "D", false);
    field(m, "yaw", "field_70177_z", "F", false);
    field(m, "pitch", "field_70125_A", "F", false);
    method(m, "getID", "func_145782_y", "()I", false);
    method(m, "getEyeHeight", "func_70047_e", "()F", false);
    method(m, "getName", "func_70005_c_", "()Ljava/lang/String;", false);
    method(m, "setSprint", "func_70031_b", "(Z)V", false);
    m = make("KeyBinding", "net.minecraft.client.settings.KeyBinding");
    method(m, "setKeyBindState", "func_74510_a", "(IZ)V", true);
    method(m, "onTick", "func_74507_a", "(I)V", true);
    method(m, "getKeyCode", "func_151463_i", "()I", false);
    m = make("GameSettings", "net.minecraft.client.settings.GameSettings");
    field(m, "keyBindAttack", "field_74312_F", "Lnet/minecraft/client/settings/KeyBinding;", false);
    m = make("Minecraft", "net.minecraft.client.Minecraft");
    field(m, "player", "field_71439_g", "Lnet/minecraft/client/entity/EntityPlayerSP;", false);
    field(m, "world", "field_71441_e", "Lnet/minecraft/client/multiplayer/WorldClient;", false);
    field(m, "gameSettings", "field_71474_y", "Lnet/minecraft/client/settings/GameSettings;", false);
    field(m, "inGameHasFocus", "field_71415_G", "Z", false);
    method(m, "getRenderViewEntity", "func_175606_aa", "()Lnet/minecraft/entity/Entity;", false);
    method(m, "getMinecraft", "func_71410_x", "()Lnet/minecraft/client/Minecraft;", true);
    m = make("EntityRenderer", "net.minecraft.client.renderer.EntityRenderer");
    field(m, "pointedEntity", "field_78528_u", "Lnet/minecraft/entity/Entity;", false);
    m = make("EntityLivingBase", "net.minecraft.entity.EntityLivingBase");
    field(m, "x", "field_70165_t", "D", false);
    field(m, "lastTickPosX", "field_70142_S", "D", false);
    field(m, "y", "field_70163_u", "D", false);
    field(m, "z", "field_70161_v", "D", false);
    field(m, "lastTickPosZ", "field_70136_U", "D", false);
    method(m, "getEyeHeight", "func_70047_e", "()F", false);
    method(m, "getID", "func_145782_y", "()I", false);
    method(m, "getName", "func_70005_c_", "()Ljava/lang/String;", false);
    m = make("EntityPlayer", "net.minecraft.entity.player.EntityPlayer");
    field(m, "x", "field_70165_t", "D", false);
    field(m, "lastTickPosX", "field_70142_S", "D", false);
    field(m, "y", "field_70163_u", "D", false);
    field(m, "z", "field_70161_v", "D", false);
    field(m, "lastTickPosZ", "field_70136_U", "D", false);
    field(m, "yaw", "field_70177_z", "F", false);
    field(m, "pitch", "field_70125_A", "F", false);
    method(m, "getEyeHeight", "func_70047_e", "()F", false);
    method(m, "getID", "func_145782_y", "()I", false);
    method(m, "getName", "func_70005_c_", "()Ljava/lang/String;", false);
    method(m, "setSprint", "func_70031_b", "(Z)V", false);
}

void Mapping::field(CM *cm, const char* name, const char* desc, bool isStatic) {
    field(cm, name, name, desc, isStatic);
}

void Mapping::method(CM *cm, const char* name, const char* desc, bool isStatic) {
    method(cm, name, name, desc, isStatic);
}

void Mapping::field(CM *cm, const char* keyName, const char* obName, const char* desc, bool isStatic) {
    std::cout << "  Mapping " << obName << " to " << keyName << std::endl;
    Mem *m = new Mem(obName, desc, isStatic);
    cm->fields.insert(std::make_pair(std::string(keyName), *m));
}

void Mapping::method(CM *cm, const char* keyName, const char* obName, const char* desc, bool isStatic) {
    std::cout << "  Mapping " << obName << desc << " to " << keyName << std::endl;
    Mem *m = new Mem(obName, desc, isStatic);
    cm->methods.insert(std::make_pair(std::string(keyName), *m));
}

CM* Mapping::make(const char* key, const char* name) {
    //struct CM cm(name);
    auto *cm = new CM(name);
    std::cout << "Mapping " << name << " to " << key << std::endl;
    lookup.insert(std::make_pair(std::string(key), cm));
    return cm;
}
